"""
Enumerate SLMs and CSMs.
"""

from .utilities import *
from time import time
from scipy.optimize import brentq, linear_sum_assignment, brute
from scipy.stats.qmc import Sobol
from scipy.spatial.transform import Rotation

np.set_printoptions(suppress=True)
Cryst = Tuple[NDArray[np.float64], NDArray[np.str_], NDArray[np.float64]]
SLM = Tuple[NDArray[np.int32], NDArray[np.int32], NDArray[np.int32]]

def cong_slm(slm: Union[SLM, NDArray[np.int32]], gA: NDArray[np.int32], gB: NDArray[np.int32]) -> tuple[SLM, int]:
    """The representative SLM of the congruence class of `slm`.

    Parameters
    ----------
    slm : slm
        `(hA, hB, q)`, representing a SLM.
    gA : (..., 3, 3) array of ints
        The rotation group of the initial crystal structure, whose elements are \
            integer matrices under fractional coordinates.
    gB : (..., 3, 3) array of ints
        The rotation group of the final crystal structure, whose elements are \
            integer matrices under fractional coordinates.

    Returns
    -------
    ss : slm
        The representative of the congruence class of `slm`.
    len_cl : int
        The size of the congruence class of `slm`.
    """
    hA, hB, q = slm
    cl = np.transpose(np.dot((gB @ hB) @ q, la.inv(gA @ hA)), axes=[2,0,1,3]).reshape(-1,9)
    cl, i = np.unique(cl.round(decimals=4), axis=0, return_index=True)
    iA, iB = np.unravel_index(i[0], (gA.shape[0], gB.shape[0]))
    hAA, qA = hnf_int(gA[iA] @ hA)
    hBB, qB = hnf_int(gB[iB] @ hB)
    ss = (hAA, hBB, qB @ q @ la.inv(qA).round().astype(int))
    return ss, len(cl)

def enumerate_slm(
    crystA: Cryst, crystB: Cryst, mu: int, kappa_max: float, tol: float = 1e-3,
    kappa: Callable[[NDArray[np.float64]], NDArray[np.float64]] = rmss,
    likelihood_ratio: float = 1e2, verbose: int = 0
) -> List[SLM]:
    """Enumerating all SLMs of multiplicity `mu` with `kappa` smaller than `kappa_max`.

    Parameters
    ----------
    crystA : cryst
        The initial crystal structure, usually obtained by `load_poscar`.
    crystB : cryst
        The final crystal structure, usually obtained by `load_poscar`.
    mu : int
        The multiplicity of SLMs to enumerate.
    kappa_max : float
        A positive threshold value of `kappa` to determine the range of singular values to generate.
    tol : float, optional
        The tolerance for determining the pure rotation group of the crystal structures. Default is 1e-3.
    kappa : callable, optional
        A function that quantifies the strain of a matrix according to its singular values. \
            Default is `rmss`, which computes the root-mean-square strain.
    likelihood_ratio : float, optional
        The expected likelihood ratio of the enumeration being complete and incomplete. Default is 1e2.
    verbose : int, optional
        The level of detail of printing. `0` means no print.

    Returns
    -------
    slmlist : list of slm
        Contains triplets of integer matrices like `(hA, hB, q)`, representing incongruent SLMs.
    """
    assert mu >= 1
    cA = crystA[0].T
    cB = crystB[0].T
    zA = crystA[2].shape[0]
    zB = crystB[2].shape[0]
    hA = hnf_list(np.lcm(zA,zB) // zA * mu)
    hB = hnf_list(np.lcm(zA,zB) // zB * mu)
    hA_inv = la.inv(hA)
    hB_inv = la.inv(hB)
    gA = get_pure_rotation(crystA, tol=tol)
    gB = get_pure_rotation(crystB, tol=tol)
    max_prob_ratio = 20
    # Compute the sampling domain of `s0`s.
    det_s = zA / zB * la.det(cB) / la.det(cA)
    if kappa.__name__ == 'rmss':
        max_strain = kappa_max * 3**0.5
    else:
        print(f"\nWarning: kappa function {kappa.__name__} is not RMSS. I hope you know what you are doing.")
        diff_kappa = lambda x: kappa(np.array([x, (det_s / x) ** 0.5, (det_s / x) ** 0.5])) - kappa_max
        a = brentq(diff_kappa, 1e-2, det_s**(1/3))
        b = brentq(diff_kappa, det_s**(1/3), 1e2)
        max_strain = max(abs(a-1), abs(b-1))
    # Enumerate SLMs.
    print(f"Enumerating SLMs (mu = {mu:d}, {kappa.__name__} <= {kappa_max:.4f}) ...")
    if verbose >= 1:
        print(f"\tprototype sampling domain: SO(3) (±{max_strain:.2f} for each matrix element)")
        print(f"\tassumed maximum probability ratio among classes: {max_prob_ratio:.1f}")
        print(f"\texpected likelihood ratio: {likelihood_ratio:.1f}")
        print("\tnum_s0\tm\tm*\telapsed_time(s)")
    slmlist = []
    iter = 0
    num_s0 = 0
    m = 0
    t = time()
    sobol_seq = Sobol(12)
    while m <= (1 + len(slmlist) * max_prob_ratio) * np.log(likelihood_ratio):    # or True for Debug!
        # Sampling `s0`s around SO(3).
        rand_num = sobol_seq.random(2)
        q0 = np.sqrt(1 - rand_num[:,0]) * np.sin(2 * np.pi * rand_num[:,1])
        q1 = np.sqrt(1 - rand_num[:,0]) * np.cos(2 * np.pi * rand_num[:,1])
        q2 = np.sqrt(rand_num[:,0]) * np.sin(2 * np.pi * rand_num[:,2])
        q3 = np.sqrt(rand_num[:,0]) * np.cos(2 * np.pi * rand_num[:,2])
        s0 = Rotation.from_quat(np.array([q0,q1,q2,q3]).T).as_matrix() + (2 * rand_num[:,3:].reshape(-1,3,3) - 1) * max_strain
        # Round `s0`s to nearest integer matrix triplets.
        q = (np.transpose(np.dot(np.dot(hB_inv, la.inv(cB) @ s0 @ cA), hA), axes=[2,3,0,1,4])).round().astype(int)
        index1 = la.det(q.reshape(-1,3,3)).round().astype(int) == 1
        # Check determinants of `q`s.
        index1 = np.nonzero(index1)[0]
        index1 = np.unravel_index(index1, (s0.shape[0], hA.shape[0], hB.shape[0]))
        # Check strains of `slm`s.
        index2 = np.nonzero(kappa(la.svd(cB @ hB[index1[2]] @ q[index1] @ hA_inv[index1[1]] @ la.inv(cA),
                                        compute_uv=False)) < kappa_max)[0]                                          
        for i in index2:
            slm = (hA[index1[1][i]], hB[index1[2][i]], q[index1[0][i], index1[1][i], index1[2][i]])
            slm, _ = cong_slm(slm, gA, gB)
            repeated = False
            for j in range(len(slmlist)):
                if (slm[0] == slmlist[j][0]).all() and (slm[1] == slmlist[j][1]).all() and (slm[2] == slmlist[j][2]).all():
                    # `slm` is repeated.
                    m = m + 1
                    repeated = True
                    break
            if not repeated:
                # `slm` is new.
                slmlist.append(slm)
                m = 0
        num_s0 += s0.shape[0]
        if verbose >= 2:
            print(f"\t{num_s0}\t{m}\t{(1 + len(slmlist) * max_prob_ratio) * np.log(likelihood_ratio):d}\t{time()-t:.2f}")
        iter = iter + 1
        if iter > 50 and len(slmlist) == 0: break
    print(f'\tFound {len(slmlist):d} incongruent SLMs (in {time()-t:.2f} seconds).')
    return slmlist

def minimize_rmsd_t(
    c: NDArray[np.float64],
    species: NDArray[np.str_],
    pA: NDArray[np.float64],
    pB: NDArray[np.float64]
) -> tuple[float, NDArray[np.int32], NDArray[np.int32]]:
    """Minimize the RMSD with fixed SLM, overall translation, variable permutation and lattice-vector translations.

    Parameters
    ----------
    c : (3, 3) array
        The matrix whose columns are cell vectors of both crystal structures.
    species : (N,) array of strs
        The array that specifies the type of each ion.
    pA, pB : (3, N) array
        The matrices whose columns are fractional coordinates of atoms.
        
    Returns
    -------
    rmsd : float
        The minimized RMSD.
    p : (Z, ) array of ints
        The permutation of the atoms in `pB` that minimizes the RMSD.
    ks : (3, Z) array of ints
        The lattice-vector translations of the atoms in `pB[:,p]` that minimizes the RMSD.
    """
    assert pA.shape[1] == pB.shape[1]
    # Determine the best lattice-vector translations for all possible atomic correspondences.
    shift = np.mgrid[-1:2, -1:2, -1:2].reshape(3,1,1,-1)
    relative_position = np.tensordot(c, (pA.reshape(3,-1,1,1) - pB.reshape(3,1,-1,1) - shift), axes=[1,0])
    d2_tensor = np.sum(relative_position ** 2, axis=0)
    d2_matrix = np.amin(d2_tensor, axis=2)
    index_shift = np.argmin(d2_tensor, axis=2)
    # Determine the best atomic correspondence.
    p = np.arange(pB.shape[1], dtype=int)
    for s in np.unique(species):
        is_s = species == s
        p[is_s] = p[is_s][linear_sum_assignment(d2_matrix[is_s][:,is_s])[1]]
    assert (np.sort(p) == np.arange(pB.shape[1])).all()
    ks = shift[:,0,0,index_shift[np.arange(pA.shape[1]),p]]
    rmsd = la.norm(c @ (pB[:,p] + ks - pA)) / np.sqrt(pA.shape[1])
    return rmsd, p, ks

def minimize_rmsd_tp(
    c: NDArray[np.float64],
    pA: NDArray[np.float64],
    pB: NDArray[np.float64]
) -> Tuple[float, NDArray[np.int32]]:
    """Minimize the RMSD with fixed SLM, overall translation, permutation, and variable lattice-vector translations.
    
    Parameters
    ----------
    c : (3, 3) array
        The matrix whose columns are cell vectors of both crystal structures.
    pA, pB : (3, N) array
        The matrices whose columns are fractional coordinates of atoms.
        
    Returns
    -------
    rmsd : float
        The minimized RMSD.
    ks : (3, Z) array of ints
        The lattice-vector translations (fractional coordinates) of the atoms in `pB` that minimizes the RMSD.
    """
    assert pA.shape[1] == pB.shape[1]
    shift = np.mgrid[-1:2, -1:2, -1:2].reshape(3,1,-1)
    relative_position = np.tensordot(c, (pA.reshape(3,-1,1) - pB.reshape(3,-1,1) - shift), axes=[1,0])
    d2_matrix = np.sum(relative_position ** 2, axis=0)
    ks = shift[:,0,np.argmin(d2_matrix, axis=1)]
    rmsd = la.norm(c @ (pB + ks - pA)) / np.sqrt(pA.shape[1])
    return rmsd, ks

def minimize_rmsd(
    crystA: Cryst, crystB: Cryst, slm: Union[SLM, NDArray[np.int32]], n_grid: int = 5
) -> tuple[float, NDArray[np.int32], NDArray[np.int32], NDArray[np.float64]]:
    """Minimize the RMSD with fixed SLM, variable permutation, overall and lattice-vector translations.

    Parameters
    ----------
    crystA : cryst
        The initial crystal structure, usually obtained by `load_poscar`.
    crystB : cryst
        The final crystal structure, usually obtained by `load_poscar`.
    slm : slm
        Triplets of integer matrices like `(hA, hB, q)`, representing a SLM.
    n_grid : int, optional
        The number of grid points for translation. Default is 5.

    Returns
    -------
    rmsd : float
        The minimum atomic displacement (RMSD) between $(S^{\\text{T}} S)^{1/4}\mathcal{A}$ and $(S^{\\text{T}} S)^{-1/4}\mathcal{B}$.
    p : (Z, ) array of ints
        The permutation of the atoms in `pB_sup` that minimizes the RMSD.
    ks : (3, Z) array of ints
        The lattice-vector translations (fractional coordinates) of the atoms in `pB_sup[:,p]` that minimizes the RMSD.
    t0 : (3, ) array of floats
        The overall translation on `pB` that minimizes the RMSD.
    """
    crystA_sup, crystB_sup, c_sup_half, f_translate = create_common_supercell(crystA, crystB, slm)
    species = crystA_sup[1]
    pA_sup = crystA_sup[2].T
    pB_sup = crystB_sup[2].T
    t0 = brute(lambda z: minimize_rmsd_t(c_sup_half, species, pA_sup, pB_sup + f_translate @ z.reshape(3,1))[0],
            ((0,1),(0,1),(0,1)), Ns=n_grid, finish=None)
    _, p, ks = minimize_rmsd_t(c_sup_half, species, pA_sup, pB_sup + f_translate @ t0.reshape(3,1))
    t0 = (pA_sup - pB_sup[:,p] - ks).mean(axis=1, keepdims=True)
    rmsd = la.norm(c_sup_half @ (pA_sup - pB_sup[:,p] - ks - t0)) / np.sqrt(pA_sup.shape[1])
    return rmsd, p, ks, t0


"""
Load/save crystal structures and CSMs from/to files.
"""

from os import makedirs
from os.path import sep, exists, splitext
import numpy as np
import numpy.linalg as la
from copy import deepcopy
from spglib import get_spacegroup, find_primitive, get_symmetry
from typing import Union, Tuple, List, Callable
from numpy.typing import NDArray, ArrayLike

np.set_printoptions(suppress=True)
Cryst = Tuple[NDArray[np.float64], NDArray[np.str_], NDArray[np.float64]]
SLM = Tuple[NDArray[np.int32], NDArray[np.int32], NDArray[np.int32]]

def load_poscar(filename: str, to_primitive: bool = True, tol: float = 1e-3, verbose: bool = True) -> Cryst:
    """Load the crystal structure from a POSCAR file.

    Parameters
    ----------
    filename : str
        The name of the POSCAR file to be read.
    to_primitive : bool, optional
        Using the primitive cell instead of the cell given by the POSCAR file. Default is True.
    tol : str
        The tolerance for `spglib` symmetry detection; default is 1e-3.

    Returns
    -------
    cryst : cryst
        The loaded crystal structure, consisting of the lattice vectors, species, and positions.
    """
    f = open(filename, mode='r')
    if verbose: print(f"Loading crystal structure '{f.readline()[:-1]}' from file '{filename}'...")
    else: f.readline()
    a = np.array(f.readline()[:-1], dtype=float)
    lattice = np.zeros((3,3), dtype=float)
    for i in range(3):
        lattice[i,:] = np.array(f.readline().split(), dtype=float)
    if la.det(lattice) < 0: lattice[2,:] = - lattice[2,:]
    lattice = a * lattice
    species_name = f.readline().split()
    species_counts = np.array(f.readline().split(), dtype=int)
    species = []
    for i in range(len(species_counts)):
        species = species + [species_name[i]] * species_counts[i]
    species = np.array(species, dtype=str)
    unit = ''
    while not unit in ['D','d','C','c','K','k']:
        unit = f.readline()[0]
    N = species_counts.sum()
    positions = np.zeros((N,3), dtype=float)
    if unit in ['D','d']:
        for i in range(N):
            positions[i,:] = np.array(f.readline().split()[:3], dtype=float)
    elif unit in ['C','c','K','k']:
        for i in range(N):
            positions[i,:] = np.dot(la.inv(lattice.transpose()), np.array(f.readline().split()[:3], dtype=float))
    f.close()
    _, numbers = np.unique(species, return_inverse=True)
    if verbose: print(f"\tSpace group: {get_spacegroup((lattice, positions, numbers), symprec=tol)}")
    if to_primitive:
        lattice, positions, numbers = find_primitive((lattice, positions, numbers), symprec=tol)
        if len(numbers) != len(species):
            if verbose: print(f"\tCell in POSCAR file is not primitive! Using primitive cell (Z = {len(numbers):d}) now.")
        else:
            if verbose: print(f"\tCell in POSCAR file is already primitive (Z = {len(numbers):d}).")
        species = np.array(species_name)[numbers]
    elif verbose: print(f"\tUsing cell in POSCAR file (Z = {len(numbers):d}).")
    cryst = (lattice, species, positions)
    return cryst

def unique_filename(message: Union[str, None], filename: str) -> str:
    """Get a unique filename by appending a number to the end of the given filename.

    Parameters
    ----------
    filename : str
        The filename to be modified.
    message : str, optional
        A message to print before the filename.

    Returns
    -------
    new_filename : str
        The modified filename with a unique number appended.
    """
    base, ext = splitext(filename)
    counter = 1
    new_filename = filename
    while exists(new_filename):
        new_filename = f"{base}-{counter}{ext}"
        counter += 1
    if message != None: print(f"{message} '{new_filename}' ...")
    return new_filename

def save_poscar(filename: Union[str, None], cryst: Cryst, crystname: Union[str, None] = None) -> None:
    """
    Save the crystal structure to a POSCAR file.

    Parameters
    ----------
    filename : str
        The name of the file to save, must not already exist in current directory. If `filename = None`, a string will be returned instead.
    cryst : cryst
        The crystal structure to be saved, consisting of the lattice vectors, species, and positions.
    crystname : str, optional
        A system description to write to the comment line of the POSCAR file. If `crystname = None`, `filename` will be used.
    """
    species_unique, species_counts = np.unique(cryst[1], return_counts=True)
    if crystname is not None: content = crystname
    else: content = ""
    content += "\n1.0\n"
    content += "\n".join(f"{v[0]:.12f}\t{v[1]:.12f}\t{v[2]:.12f}" for v in cryst[0].tolist())
    content += "\n" + " ".join(species_unique.tolist())
    content += "\n" + " ".join(str(n) for n in species_counts.tolist())
    content += "\nDirect\n"
    content += "\n".join(f"{p[0]:.12f}\t{p[1]:.12f}\t{p[2]:.12f}" for p in cryst[2].tolist())
    if filename is not None:
        f = open(filename, mode='x')
        f.write(content)
        f.close()
        return
    else:
        return content

def check_chem_comp(speciesA, speciesB):
    spA, ctA = np.unique(speciesA, return_counts=True)
    spB, ctB = np.unique(speciesB, return_counts=True)
    assert (spA == spB).all()
    assert np.dot(ctA, ctA) * np.dot(ctB, ctB) == np.dot(ctA, ctB) ** 2
    return

def create_common_supercell(crystA: Cryst, crystB: Cryst, slm: SLM) -> Tuple[Cryst, Cryst, NDArray[np.float64], NDArray[np.float64]]:
    """Create supercell structures representing $\mathcal{A}$ (initial structure)and $\sqrt{S^{\\text{T}} S}S^{-1}\mathcal{B}$ \
        (rotation-free final structure). Also return the half-distorted supercell and translation cell.
    
    Parameters
    ----------
    crystA, crystB : cryst
        The initial and final structures.
    slm : slm
        The SLM of the CSM.
    
    Returns
    -------
    crystA_sup : cryst
        The supercell of $\mathcal{A}$.
    crystB_sup_final : cryst
        The supercell of $\sqrt{S^{\\text{T}} S}\mathcal{A}$, which equals the supercell of $(S^{\\text{T}} S)^{-1/2}\mathcal{B}$.
    c_sup_half : (3, 3) array of floats
        The half-distorted supercell, which equals the supercel of $(S^{\\text{T}} S)^{1/4}\mathcal{A}$ and that of $(S^{\\text{T}} S)^{-1/4}\mathcal{B}$.
    f_translate : (3, 3) array of floats
        The translation cell (fractional coordinates) of the shuffle whose lattice is generated by`c_sup_half`.
    """
    # Unpacking crystal structures.
    cA = crystA[0].T
    cB = crystB[0].T
    speciesA = crystA[1]
    speciesB = crystB[1]
    pA = crystA[2].T
    pB = crystB[2].T
    check_chem_comp(speciesA, speciesB)
    
    # Determining the supercell geometries from the SLM.
    hA, hB, q = slm
    deform = cB @ hB @ q @ la.inv(cA @ hA)
    u, sigma, vT = la.svd(deform)
    c_sup_half, q_sup = niggli_cell(vT.T @ np.diag(sigma ** 0.5) @ vT @ cA @ hA)         # The half-distorted supercell.
    mA = hA @ q_sup
    mB = hB @ q @ q_sup
    cA_sup = cA @ mA
    cB_sup_final = (u @ vT).T @ cB @ mB                     # The rotation-free orientation of `crystB_sup`.
    
    # Sorting supercell species and positions.
    speciesA_sup = np.tile(speciesA, la.det(mA).round().astype(int))
    speciesB_sup = np.tile(speciesB, la.det(mB).round().astype(int))
    pA_sup = la.inv(mA) @ (pA.reshape(3,1,-1) + int_vec_inside(mA).reshape(3,-1,1)).reshape(3,-1)
    pB_sup = la.inv(mB) @ (pB.reshape(3,1,-1) + int_vec_inside(mB).reshape(3,-1,1)).reshape(3,-1)
    argsortA = np.argsort(speciesA_sup)
    argsortB = np.argsort(speciesB_sup)
    assert (speciesA_sup[argsortA] == speciesB_sup[argsortB]).all(), "Please report this bug to wfc@pku.edu.cn if you see this message."
    species_sup = speciesA_sup[argsortA]
    pA_sup = pA_sup[:,argsortA]
    pB_sup = pB_sup[:,argsortB]
    
    # Computing output.
    crystA_sup = (cA_sup.T, species_sup, pA_sup.T)
    crystB_sup_final = (cB_sup_final.T, species_sup, pB_sup.T)
    f_translate = niggli_cell(matrix_gcd(la.inv(mA), la.inv(mB)))[0]
    return crystA_sup, crystB_sup_final, c_sup_half, f_translate

def int_arrays_to_pair(crystA: Cryst, crystB: Cryst, slm: SLM,
    p: NDArray[np.int32], ks: NDArray[np.int32], centered: bool = True
) -> Tuple[Cryst, Cryst]:
    """Convert the integer arrays representation `(slm, p, translations)` of a CSM to a pair of crysts.
    
    Parameters
    ----------
    crystA, crystB : cryst
        The initial and final structures.
    slm : slm
        The SLM of the CSM.
    p : (Z, ) array of ints
        The permutaion of the shuffle.
    ks : (3, Z) array of ints
        The lattice-vector translations of the shuffle.
    centered : bool, optional
        Whether to make the centers of `crystA_sup` and `crystB_sup_final` coincide. Default is True.
        
    Returns
    -------
    crystA_sup : cryst
        The initial structure, whose lattice vectors and atoms are matched to `crystB_sup` according to the CSM.
    crystB_sup_final : cryst
        The final structure, whose lattice vectors and atoms are matched to `crystA_sup` according to the CSM, with rotation-free orientation.
    """
    crystA_sup, crystB_sup, _, _ = create_common_supercell(crystA, crystB, slm)
    pA_sup = crystA_sup[2].T
    pB_sup = crystB_sup[2].T[:,p] + ks
    if centered:
        pB_sup = pB_sup - np.mean(pB_sup - pA_sup, axis=1, keepdims=True)
    return crystA_sup, (crystB_sup[0], crystB_sup[1][p], pB_sup.T)

def rmss(x: NDArray[np.float64]) -> NDArray[np.float64]:
    """Root-mean-square strain of given singular values.

    Parameters
    ----------
    x : (..., 3) array
        The singular values of 3*3 matrices.
    
    Returns
    -------
    rms_strain : (...) array
        Root-mean-square of `x - 1`.
    """
    return np.sqrt(np.mean((x - 1) ** 2, axis=-1))

def get_pure_rotation(cryst: Cryst, tol: float = 1e-3) -> NDArray[np.int32]:
    """Find all pure rotations appeared in the space group of `cryst`.

    Parameters
    ----------
    cryst : 3-tuple
        `(lattice, species, positions)`, representing the crystal structure, usually obtained by `load_poscar`.
    tol : float, optional
        The tolerance for `spglib` symmetry detection; default is 1e-3.
    
    Returns
    -------
    g : (..., 3, 3) array of ints
        A point group of the first kind, containing all pure rotations appeared in the space group of `cryst`, \
            elements of which are integer matrices (under fractional coordinates).
    """
    species = cryst[1]
    n = len(species)
    numbers = np.zeros(n, dtype=int)
    temp_s = ''
    temp_n = 0
    for i in range(n):
        if species[i] != temp_s:
            temp_n = temp_n + 1
            temp_s = species[i]
        numbers[i] = temp_n
    g = get_symmetry((cryst[0],cryst[2],numbers), symprec=tol)['rotations']
    g = g[la.det(g).round(decimals=4)==1,:,:]
    g = np.unique(g, axis=0)
    return g

def int_vec_inside(c: NDArray[np.int32]) -> NDArray[np.int32]:
    """Integer vectors inside the cell `c` whose elements are integers.

    Parameters
    ----------
    c : (3, 3) array of ints
        A matrix whose columns are integer cell vectors.
    
    Returns
    -------
    v_int : (3, ...) array of ints
        Its columns are vectors satisfying `v = c @ k`, where `k[0]`, `k[1]`, `k[2]` $\in [0, 1)$.
    """
    assert c.dtype == int
    vertices = c @ np.mgrid[0:2,0:2,0:2].reshape(3,-1)
    candidates = np.mgrid[np.amin(vertices[0,:]):np.amax(vertices[0,:])+1, np.amin(vertices[1,:]):np.amax(vertices[1,:])+1, \
        np.amin(vertices[2,:]):np.amax(vertices[2,:])+1].reshape(3,-1)
    fractional = (la.inv(c) @ candidates).round(decimals=7)
    is_inside = (np.prod(fractional < 1, axis=0) * np.prod(fractional >= 0, axis=0)).astype(bool)
    assert np.sum(is_inside) == la.det(c).round().astype(int)
    return candidates[:,is_inside]

def matrix_gcd(m1: ArrayLike, m2: ArrayLike, max_divisor = 10000) -> NDArray[np.float64]:
    """Return a greatest common divisor of rational matrices `m1` and `m2`.
    
    Parameters
    ----------
    m1, m2 : (3, 3) array_like
        Nonsingular rational matrices.
    max_divisor : int
        A positive integer. The least common multiple of all divisors in `m` should not be greater than `max_divisor`.
    
    Returns
    -------
    d : (3, 3) array
        The greatest common divisor of `m1` and `m2` in Hermite normal form.
    """
    assert la.det(m1) != 0 and la.det(m2) != 0
    d = hnf_rational(np.hstack((m1, m2)), max_divisor=max_divisor)[:,:3]
    if m1.dtype == int and m2.dtype == int: d = d.round().astype(int)
    return d

def matrix_lcm(m1: ArrayLike, m2: ArrayLike) -> NDArray[np.int32]:
    """Return a least common multiple of integer matrices `m1` and `m2`.
    
    Parameters
    ----------
    m1, m2 : (3, 3) array_like
        Nonsingular integer matrices.
    
    Returns
    -------
    m : (3, 3) array
        The least common multiple of `m1` and `m2` in Hermite normal form.
    """
    assert m1.dtype == int and m2.dtype == int
    assert (la.det([m1, m2]) != 0).all()
    h = hnf_rational(np.hstack((la.inv(m1.T), la.inv(m2.T))))[:,:3]
    m = la.inv(h.T).round().astype(int)
    return m

def int_fact(n: int) -> List[Tuple[int, int]]:
    """Factorize positive integer `n` into products of two integers.

    Parameters
    ----------
    n : int
        The integer to be factorized.
    
    Returns
    -------
    l : list of 2-tuples of ints
        Contains all `(a, b)` such that a*b=n.
    """
    l = []
    for a in range(1,n+1):
        if n % a == 0: l.append((a, n//a))
    return l

def hnf_list(det: int) -> NDArray[np.int32]:
    """Enumerate all 3*3 column Hermite normal forms (HNFs) with given determinant.

    Parameters
    ----------
    det : int
        The determinant of HNFs.
    
    Returns
    -------
    l : (..., 3, 3) array of ints
        Contains all HNFs with determinant `det`.
    """
    # Enumerate 3-factorizations of `det`.
    diag_list = []
    for a, aa in int_fact(det):
        for b, c in int_fact(aa):
            diag_list.append((a, b, c))
    # Enumerate HNFs.
    l = []
    for diag in diag_list:
        for h21 in range(diag[1]):
            for h31 in range(diag[2]):
                for h32 in range(diag[2]):
                    h = np.diag(diag)
                    h[1,0] = h21
                    h[2,0] = h31
                    h[2,1] = h32
                    l.append(h)
    l = np.array(l, dtype=int)
    return l

def hnf_int(m: NDArray[np.int32]) -> tuple[NDArray[np.int32], NDArray[np.int32]]:
    """Decompose square integer matrix `m` into product of HNF matrix `h` and unimodular matrix `q`.

    Parameters
    ----------
    m : (N, N) array of ints
        The integer matrix to decompose, with positive determinant.
    
    Returns
    -------
    h : (N, N) array of ints
        The column-style Hermite normal form of `m`.
    q : (N, N) array of ints
        The unimodular matrix satisfying `m` = `h @ q`.
    """
    assert m.dtype == int and la.det(m) > 0
    N = m.shape[0]
    h = deepcopy(m)
    for i in range(N):
        while not (h[i,i+1:] == 0).all():
            col_nonzero = i + np.nonzero(h[i,i:])[0]
            i0 = col_nonzero[np.argpartition(np.abs(h[i,col_nonzero]), kth=0)[0]]
            h[:,[i,i0]] = h[:,[i0,i]]
            if h[i,i] < 0: h[:,i] = - h[:,i]
            h[:,i+1:] = h[:,i+1:] - np.outer(h[:,i], (h[i,i+1:] / h[i,i]).round().astype(int))
        if h[i,i] < 0: h[:,i] = - h[:,i]
        h[:,:i] = h[:,:i] - np.outer(h[:,i], h[i,:i] // h[i,i])
    q = (la.inv(h) @ m).round().astype(int)
    return h, q

def hnf_rational(m: ArrayLike, max_divisor = 10000) -> NDArray[np.float64]:
    """The Hermite normal form (HNF) of full-row-rank rational matrix `m` (not necessarily square or integer).
    
    Parameters
    ----------
    m : (M, N) array_like, M <= N
        The full-row-rank rational matrix to reduce.
    max_divisor : int
        A positive integer. The least common multiple of all divisors in `m` should not be greater than `max_divisor`.
    
    Returns
    -------
    h : (M, N) array
        The HNF of `m` obtained via elementary column operations over integers.
    """
    for divisor in range(1, max_divisor+1):
        if (np.absolute(np.rint(m * divisor) - m * divisor) <= 1 / max(10000,max_divisor)).all(): break
        elif divisor == max_divisor: raise ValueError("Input matrix must be rational.")
    h = (m * divisor).round().astype(int)
    M, N = h.shape
    assert M <= N and la.matrix_rank(h, tol=1/max(10000,max_divisor)) == M
    for i in range(M):
        while not (h[i,i+1:] == 0).all():
            col_nonzero = i + np.nonzero(h[i,i:])[0]
            i0 = col_nonzero[np.argpartition(np.abs(h[i,col_nonzero]), kth=0)[0]]
            h[:,[i,i0]] = h[:,[i0,i]]
            if h[i,i] < 0: h[:,i] = - h[:,i]
            h[:,i+1:] = h[:,i+1:] - np.outer(h[:,i], (h[i,i+1:] / h[i,i]).round().astype(int))
        if h[i,i] < 0: h[:,i] = - h[:,i]
        h[:,:i] = h[:,:i] - np.outer(h[:,i], h[i,:i] // h[i,i])
    return h / divisor

def vector_reduce(v: NDArray, divisors: NDArray) -> NDArray:
    """Minimizing the norm of `v` by adding and subtracting columns of `divisors`.

    Parameters
    ----------
    v : (N,) array
        The vector to be reduced.
    divisors : (N, ...) array
        The vectors used to translate `v`.
    
    Returns
    -------
    vv : (N,) array
        The reduced `v` with minimum Euclidean norm.
    """
    vv = deepcopy(v)
    converged = False
    while not converged:
        converged = True
        for i in range(divisors.shape[1]):
            v0 = divisors[:,i]
            while la.norm(vv + v0) < la.norm(vv):
                converged = False
                vv = vv + v0
            while la.norm(vv - v0) < la.norm(vv):
                converged = False
                vv = vv - v0
    return vv

def niggli_cell(c: NDArray[np.float64]) -> tuple[NDArray[np.float64], NDArray[np.int32]]:
    """Reduce cell `c` to its Niggli cell.

    Parameters
    ----------
    c : (3, 3) array
        The cell to be reduced, whose columns are cell vectors.
    
    Returns
    -------
    cc : (3, 3) array
        The Niggli cell, with shortest right-handed cell vectors.
    q : (3, 3) array of ints
        The unimodular matrix satisfying `cc = c @ q`.
    """
    c0 = np.zeros((3,3))
    cc = deepcopy(c)
    while (cc != c0).any():
        c0 = deepcopy(cc)
        cc = cc[:,np.argsort(la.norm(cc, axis=0))]
        cc[:,2] = vector_reduce(cc[:,2], cc[:,0:2])
    if la.det(cc) < 0: cc = -cc
    q = (la.inv(c) @ cc).round().astype(int)
    return cc, q


import os
import glob
import shutil
import subprocess

sep = os.path.sep

def clean_ext(ext_list):
    for ext in ext_list:
        for file in glob.glob("*." + ext):
            os.remove(file)

def clean_csm():
    for item in os.listdir("."):
        if os.path.isdir(item) and item.startswith("CSM_"):
            shutil.rmtree(item)

def test_version():
    result = subprocess.run(
        ["crystmatch", "-v"], 
        capture_output=True, text=True, timeout=10
    )
    assert result.returncode == 0
    assert "crystmatch" in result.stdout

def test_help():
    result = subprocess.run(
        ["crystmatch", "-h"], 
        capture_output=True, text=True, timeout=5
    )
    assert result.returncode == 0
    assert "wfc@pku.edu.cn" in result.stdout

def test_enumeration():
    clean_ext(["npz", "pdf", "csv"])
    result = subprocess.run(
        ["crystmatch"], input=f"enumeration\n3\n0.1\ntest{sep}wz.txt\ntest{sep}zb.txt\n",
        capture_output=True, text=True, timeout=10
    )
    assert result.returncode == 0
    assert "Found 0" in result.stdout
    assert "Found 2" in result.stdout
    assert "Found 10" in result.stdout
    result = subprocess.run(
        ["crystmatch", "-E", "3", "0.1", "-I", f"test{sep}wz.txt", "-F", f"test{sep}zb.txt"],
        capture_output=True, text=True, timeout=10
    )
    assert result.returncode == 0
    assert "A total of 12" in result.stdout
    assert "'CSM_LIST-m3s0.10-1.npz'" in result.stdout
    assert os.path.exists("CSM_LIST-m3s0.10-1.npz")
    assert os.path.exists("PLOT-m3s0.10-1.pdf")
    assert os.path.exists("TABLE-m3s0.10-1.csv")

def test_read():
    clean_csm()
    result = subprocess.run(
        ["crystmatch"], input=f"read\nCSM_LIST-m3s0.10.npz\n7 11\n",
        capture_output=True, text=True, timeout=5
    )
    assert result.returncode == 0
    result = subprocess.run(
        ["crystmatch", "-R", "CSM_LIST-m3s0.10.npz", "-e", "7", "11"],
        capture_output=True, text=True, timeout=5
    )
    assert result.returncode == 0
    assert os.path.exists(f"CSM_11-1{sep}POSCAR_I")

def test_single():
    result = subprocess.run(
        ["crystmatch"], input=f"single\nCSM_7{sep}POSCAR_I\nCSM_7{sep}POSCAR_F\n",
        capture_output=True, text=True, timeout=5
    )
    assert result.returncode == 0
    assert "rmss = 0.0842" in result.stdout
    assert "rmsd = 1.1126" in result.stdout
    result = subprocess.run(
        ["crystmatch", "-S", "-I", f"CSM_11{sep}POSCAR_I", "-F", f"CSM_11{sep}POSCAR_F"],
        capture_output=True, text=True, timeout=5
    )
    assert result.returncode == 0
    assert os.path.exists(f"CSM_single-1{sep}POSCAR_I")
    assert os.path.exists(f"CSM_single-1{sep}POSCAR_F-optimized")
    
def test_clean():
    clean_ext(["npz", "pdf", "csv"])
    clean_csm()
    assert 1 + 1 == 2


"""
Analyze and visualize CSMs.
"""

from .utilities import *
from matplotlib import rcParams, colors
import matplotlib.pyplot as plt
rcParams.update({
    'font.family': 'serif',
    'pgf.rcfonts': False,
    'figure.dpi': 150
})

np.set_printoptions(suppress=True)
Cryst = Tuple[NDArray[np.float64], NDArray[np.str_], NDArray[np.float64]]
SLM = Tuple[NDArray[np.int32], NDArray[np.int32], NDArray[np.int32]]

def multiplicity(crystA: Cryst, crystB: Cryst, slmlist: Union[SLM, List[SLM], NDArray[np.int32]]) -> Union[int, NDArray[np.int32]]:
    """Return multiplicities of elements in `slmlist`.

    Parameters
    ----------
    crystA : cryst
        The initial crystal structure, usually obtained by `load_poscar`.
    crystB : cryst
        The final crystal structure, usually obtained by `load_poscar`.
    slmlist : list of slm
        A list of SLMs, each represented by a triplet of integer matrices like `(hA, hB, q)`.

    Returns
    -------
    mu : int or (...,) array of ints
        Multiplicities of each SLM in `slmlist`.
    """
    slmlist = np.array(slmlist)
    zA = crystA[2].shape[0]
    zB = crystB[2].shape[0]
    dA = np.lcm(zA,zB) // zA
    if len(slmlist.shape) == 3:
        return la.det(slmlist[0]).round().astype(int) // dA
    else:
        return la.det(slmlist[:,0,:,:]).round().astype(int) // dA

def sing_val(crystA: Cryst, crystB: Cryst, slmlist: Union[List[SLM], NDArray[np.int32]]) -> NDArray[np.float64]:
    """Return singular values of elements in `slmlist`.

    Parameters
    ----------
    crystA : cryst
        The initial crystal structure, usually obtained by `load_poscar`.
    crystB : cryst
        The final crystal structure, usually obtained by `load_poscar`.
    slmlist : list of slm
        A list of SLMs, each represented by a triplet of integer matrices like `(hA, hB, q)`.

    Returns
    -------
    sv : (..., 3) array
        Contains singular values of each SLM in `slmlist`.
    """
    cA = crystA[0].T
    cB = crystB[0].T
    hA, hB, q = zip(*slmlist)
    deform = cB @ np.array(hB) @ np.array(q) @ la.inv(cA @ np.array(hA))
    sv = la.svd(deform, compute_uv=False)
    return sv

def deform_distance(slmlist: ArrayLike, s0: ArrayLike, crystA: Cryst, crystB: Cryst) -> NDArray[np.float64]:
    """The Frobenius distance between deformation gradients.

    Parameters
    ----------
    slmlist : list of slm
        A list of SLMs, each represented by a triplet of integer matrices like `(hA, hB, q)`.
    s0 : slm
        `(hA, hB, q)`, representing a SLM.
    crystA, crystB : cryst
        `(lattice, species, positions)`, representing the crystal structure, usually obtained by `load_poscar`.
    
    Returns
    -------
    dlist : (...,) array
        Contains Frobenius distances from `slmlist` to `s0`, where equivalent SLMs coincide.
    """
    cA = crystA[0].T
    cB = crystB[0].T
    gA = get_pure_rotation(crystA)
    gB = get_pure_rotation(crystB)
    hA0, hB0, q0 = s0
    x0 = np.transpose(np.dot((gB @ hB0) @ q0, la.inv(gA @ hA0)), axes=[2,0,1,3]).reshape(-1,9)
    _, i = np.unique(x0.round(decimals=4), axis=0, return_index=True)
    cl0 = cB @ x0[i,:].reshape(-1,3,3) @ la.inv(cA)
    ss = cB @ slmlist[:,1,:,:] @ slmlist[:,2,:,:] @ la.inv(cA @ slmlist[:,0,:,:])
    dlist = np.amin(la.norm(ss.reshape(-1,1,9), cl0.reshape(1,-1,9), axis=2),axis=1)
    return dlist

def orient_matrix(vi: ArrayLike, vf: ArrayLike, wi: ArrayLike, wf: ArrayLike) -> NDArray[np.float64]:
    """Rotation matrix `r` such that `r @ vi` || `vf` and `r @ wi` || `wf`.

    Parameters
    ----------
    vi, vf : (3,) array_like
        Vectors (cartesian coordinates) that satisfy `r @ vi` || `vf`.
    wi, wf : (3,) array_like
        Vectors (cartesian coordinates) that satisfy `r @ wi` || `wf`.
        
    Returns
    -------
    r : (3, 3) array
        A rotation matrix representing the given orientation relationship.
    """
    b = np.array([vi, np.cross(vi, wi), np.cross(vi, np.cross(vi, wi))]).T
    c = np.array([vf, np.cross(vf, wf), np.cross(vf, np.cross(vf, wf))]).T
    b = b * la.norm(b, axis=0).reshape(1,3) ** -1
    c = c * la.norm(c, axis=0).reshape(1,3) ** -1
    r = c @ b.T
    return r

def deviation_angle(
    crystA: Cryst,
    crystB: Cryst,
    slmlist: Union[List[SLM], NDArray[np.int32]],
    r: NDArray[np.float64],
    uspfix: bool = False
) -> NDArray[np.float64]:
    """Calculate how much each SLM in `slmlist` differ from a given orientation relationship.

    Parameters
    ----------
    crystA : cryst
        The initial crystal structure, usually obtained by `load_poscar`.
    crystB : cryst
        The final crystal structure, usually obtained by `load_poscar`.
    slmlist : list of slm
        A list of SLMs, each represented by a triplet of integer matrices like `(hA, hB, q)`.
    r : (3, 3) array
        A rotation matrix representing the given orientation relationship.
    uspfix : bool, optional
        Whether to fix the uniformed scaled plane. Default is False.

    Returns
    -------
    anglelist : (...,) array
        Contains rotation angles that measure the difference of each SLM and the given orientation.
    """
    assert la.det(r).round(decimals=4) == 1
    cA = crystA[0].T
    cB = crystB[0].T
    rA = cA @ get_pure_rotation(crystA) @ la.inv(cA)
    rB = cB @ get_pure_rotation(crystB) @ la.inv(cB)
    r_equiv = np.transpose(np.dot(np.dot(rB, r), rA), axes=(2,0,1,3)).reshape(-1,3,3)
    hA, hB, q = zip(*slmlist)
    s = cB @ np.array(hB) @ np.array(q) @ la.inv(cA @ np.array(hA))
    u, sigma, vT = la.svd(s)
    rS = u @ vT
    if uspfix:
        eps = np.array([[[0,0,0],[0,0,-1],[0,1,0]],[[0,0,1],[0,0,0],[-1,0,0]],[[0,-1,0],[1,0,0],[0,0,0]]])
        v_cross = np.tensordot(vT[:,1,:], eps, axes=(1,0))
        s1 = sigma[:,0]
        s2 = sigma[:,1]
        s3 = sigma[:,2]
        theta = np.arctan(np.sqrt((s1**2 - s2**2) * (s2**2 - s3**2)) / (s1 * s3 + s2**2))
        rH = np.eye(3).reshape(1,3,3) + np.sin(theta).reshape(-1,1,1) * v_cross + (1 - np.cos(theta)).reshape(-1,1,1) * (v_cross @ v_cross)
        rS = np.array([rS @ rH, rS @ la.inv(rH)])       # rS.shape = (2, ..., 3, 3)
        anglelist = np.arccos(np.clip(0.5 * (-1 + np.amax(np.trace(np.dot(la.inv(r_equiv), rS), axis1=1, axis2=4), axis=(0,1))), -1, 1))
    else:
        anglelist = np.arccos(np.clip(0.5 * (-1 + np.amax(np.trace(np.dot(la.inv(r_equiv), rS), axis1=1, axis2=3), axis=0)), -1, 1))
    return anglelist.round(decimals=7)

def save_scatter(
    filename : str,
    rmsslist: ArrayLike,
    rmsdlist: ArrayLike,
    colorlist: ArrayLike,
    cmap : colors.Colormap = plt.cm.get_cmap('viridis'),
    cbarlabel: str = None
) -> None:
    """Scatter plot of the CSMs with colorbar.

    Parameters
    ----------
    filename : str
        The filename of the saved plot.
    rmsslist : (N,) array_like
        The root-mean-square strain of each CSM.
    rmsdlist : (N,) array_like
        The root-mean-square distance of each CSM.
    colorlist : (N,) array_like
        Some quantity of each CSM, which is to be colored.
    cmap : `matplotlib.colors.Colormap`, optional
        The colormap to use. Default is `plt.cm.get_cmap('viridis')`.
    cbarlabel : str, optional
        The label of the colorbar. Default is None, in which case the filename is used.
    """
    rmsslist = np.array(rmsslist)
    rmsdlist = np.array(rmsdlist)
    colorlist = np.array(colorlist)
    plt.figure()
    ax = plt.subplot()
    ind = np.argsort(colorlist)[::-1]
    sc = plt.scatter(rmsslist[ind], rmsdlist[ind], marker='d', c=colorlist[ind], cmap=cmap, s=20)
    ind0 = colorlist==0
    n0 = np.sum(ind0)
    if n0 >= 1:
        print(f"\nThere are {n0:d} CSMs (indices: {', '.join(np.nonzero(ind0)[0].astype(str).tolist())}) with {cbarlabel}=0, emphasized by pink stars in the plot.")
        plt.scatter(rmsslist[ind0], rmsdlist[ind0], marker='*', color=(1.0,0.75,0.95), s=12)
    plt.xlabel("Root-mean-square strain (RMSS)", fontsize=15)
    plt.ylabel("RMSD / Å", fontsize=15)
    plt.xlim(0, np.amax(rmsslist) * 1.05)
    plt.ylim(min(0, np.amin(rmsdlist) - 0.1), np.amax(rmsdlist) + 0.1)
    if colorlist.dtype == int:
        cbar = plt.colorbar(sc, aspect=40, ticks=np.unique(colorlist))
    else:
        cbar = plt.colorbar(sc, aspect=40)
    cbar.set_label(cbarlabel if cbarlabel != None else filename.split('.')[0], fontsize=13)
    ax.tick_params(axis='both', which='major', labelsize=11)
    ax.tick_params(axis='both', which='minor', labelsize=8)
    plt.savefig(f"{filename}", bbox_inches='tight')
    return


from .utilities import *
from .enumeration import *
from .analysis import *
import sys
import argparse

sys.stdout = open(sys.stdout.fileno(), mode='w', buffering=1, encoding='utf-8', closefd=False)
sys.stderr = open(sys.stderr.fileno(), mode='w', buffering=1, encoding='utf-8', closefd=False)

__name__ = "crystmatch"
__version__ = "1.1.2"
__author__ = "Fang-Cheng Wang"
__email__ = "wfc@pku.edu.cn"
__description__ = 'Enumerating and analyzing crystal-structure matches for solid-solid phase transitions.'
__url__ = 'https://github.com/Fangcheng-Wang/crystmatch'
__epilog__ = 'The current version (v' + __version__ + ') may contain bugs. To get the latest version, please see: \
\n\n\t' + __url__ + '\n\nIf you use crystmatch in your research, please cite the following paper: \n\n\t\
[1] FC Wang, QJ Ye, YC Zhu, and XZ Li, Physical Review Letters 132, 086101 (2024) (https://arxiv.org/abs/2305.05278)\n\n\
You are also welcome to contact me at ' + __email__ + ' for any questions or comments.'

def enum_rep(crystA: Cryst, crystB: Cryst, mu_max: int, rmss_max: float, tol: float):
    
    # enumerating SLMs
    print(f"\nEnumerating incongruent SLMs for mu <= {mu_max} and rmss <= {rmss_max:.4f} ...")
    slmlist = []
    for mu in range(1,mu_max+1): slmlist = slmlist + enumerate_slm(crystA, crystB, mu, rmss_max, tol=tol)
    print(f"\nA total of {len(slmlist):d} incongruent SLMs are enumerated.")
    if len(slmlist) == 0: raise Warning("No SLM is found. Try larger arguments for '--enumeration' or check if the input POSCARs are correct.")
    
    slmlist = np.array(slmlist)
    _, ind = np.unique((slmlist[:,1,:,:] @ slmlist[:,2,:,:] @ la.inv(slmlist[:,0,:,:])).round(decimals=4), axis=0, return_index=True)
    slmlist = slmlist[ind]
    mulist = multiplicity(crystA, crystB, slmlist)
    rmsslist = rmss(sing_val(crystA, crystB, slmlist))
    ind = np.lexsort((rmsslist.round(decimals=4), mulist))
    slmlist = slmlist[ind]
    mulist = mulist[ind]
    rmsslist = rmsslist[ind]
    print(f"They have {len(slmlist):d} distinct deformation gradients, and thus {len(slmlist):d} representative CSMs.")

    # computing representative CSMs
    csm_bins = [np.array("arr_mu.npy saves the shuffles of those CSMs with multiplicity mu", dtype=str)]
    rmsdlist = np.inf * np.ones(len(slmlist))
    print(f"\nOptimizing RMSDs for {slmlist.shape[0]} shuffles ...")
    time0 = time()
    zlcm = np.lcm(len(crystA[1]), len(crystB[1]))
    for mu in range(1,mu_max+1):
        n_csm = np.sum(mulist == mu)
        if n_csm == 0:
            csm_bins.append(np.array(f"No CSM with multiplicity {mu}", dtype=str))
            continue
        shufflelist = np.zeros((n_csm, mu * zlcm, 4), dtype=int)
        for i in range(n_csm):
            d, p, ks, _ = minimize_rmsd(crystA, crystB, slmlist[mulist == mu][i])
            rmsdlist[np.sum(mulist < mu) + i] = d
            shufflelist[i,:,0] = p
            shufflelist[i,:,1:] = ks.T
        ind = np.lexsort((rmsdlist[mulist == mu].round(decimals=4), rmsslist[mulist == mu].round(decimals=4)))
        slmlist[mulist == mu] = slmlist[mulist == mu][ind]
        mulist[mulist == mu] = mulist[mulist == mu][ind]
        rmsslist[mulist == mu] = rmsslist[mulist == mu][ind]
        rmsdlist[mulist == mu] = rmsdlist[mulist == mu][ind]
        shufflelist = shufflelist[ind]
        csm_bins.append(shufflelist)
    print(f"\tObtained representative CSMs and their RMSDs (in {time()-time0:.2f} seconds).")
    
    return csm_bins, slmlist, mulist, rmsslist, rmsdlist

def main():
    time0 = time()
    parser = argparse.ArgumentParser(
        prog=__name__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        usage='%(prog)s [-h] [-initial POSCAR_I] [-final POSCAR_F] [mode options] [output options]',
        description=__description__, epilog=__epilog__)
    parser.add_argument("-v", "--version", action='version', version="%(prog)s " + __version__)
    parser.add_argument("-E", "--enumeration", nargs=2, type=float, metavar=('MAX_MU','MAX_RMSS'),
                        help="use 'enumeration' mode, with MAX_MU and MAX_RMSS as the multiplicity and RMSS upper bounds")
    parser.add_argument("-R", "--read", type=str, metavar='CSM_LIST',
                        help="use 'read' mode, with CSMs loaded from CSM_LIST (an NPZ file generated by crystmatch)")
    parser.add_argument("-S", "--single", action='store_true', default=None,
                        help="use 'single-CSM' mode, with POSCARs specified by -I and -F to determine a single CSM")
    parser.add_argument("-I", "--initial", type=str, metavar='POSCAR_I',
                        help="POSCAR file of the initial crystal structure")
    parser.add_argument("-F", "--final", type=str, metavar='POSCAR_F',
                        help="POSCAR file of the final crystal structure")
    parser.add_argument("-t", "--tolerance", type=float, default=1e-3, metavar='TOL',
                        help="tolerance for determining crystal symmetry; default is 1e-3")
    parser.add_argument("-e", "--export", nargs='+', type=int, metavar=('index1', 'index2'),
                        help="export CSMs from CSM_LIST with the given indices")
    parser.add_argument("-o", "--orientation", nargs=12, type=float, metavar=('vix','viy','viz','vfx','vfy','vfz','wix','wiy','wiz','wfx','wfy','wfz'),
                        help="benchmark CSMs by the orientation relationship: vi||vf, wi||wf")
    parser.add_argument("-c", "--csv", action='store_true', help="whether to create CSV file")
    parser.add_argument("-p", "--plot", action='store_true', help="whether to create scatter plot")
    parser.add_argument("-u", "--uspfix", action='store_true', help="use USP-fixed manner [1] in '--orientation'")
    parser.add_argument("-A", nargs='*', type=str, metavar='', help="deprecated")
    parser.add_argument("-B", nargs='*', type=str, metavar='', help="deprecated")
    args = parser.parse_args()

    if args.A is not None or args.B is not None:
        raise DeprecationWarning("'-A -B' is deprecated since version 1.1.0, use '-I -F' instead. See 'crystmatch -h' for usage.")
        
    if sum(1 for x in [args.enumeration, args.read, args.single] if x is not None) >= 2:
        raise ValueError("Cannot choose more than one mode.")

    if sum(1 for x in [args.enumeration, args.read, args.single] if x is not None) == 0:
        print("No mode is specified (not recommended, see 'crystmatch -h' for common usage).")
        mode = input("Please enter a mode ('enumeration' / 'read' / 'single'): ")
        if mode in ["'enumeration'", 'enumeration', 'e', 'E']:
            mu_max = int(input("Enter the maximum multiplicity: "))
            rmss_max = float(input("Enter the maximum root-mean-square-strain: "))
            args.enumeration = (mu_max, rmss_max)
        elif mode in ["'read'", 'read', 'r', 'R']:
            args.read = input("Enter the path of the CSM_LIST file: ")
            args.export = np.array(input("Enter the indices of CSMs to be exported (separated by space), or leave blank for no export: ").split(), dtype=int)
            if args.export.shape[0] == 0: args.export = None
        elif mode in ["'single'", 'single', 'single-CSM', 's', 'S']:
            args.single = True
            print("Warning: The CSM will be uniquely determined by the initial and final POSCARs. I hope you know what you are doing.")
        else:
            raise ValueError(f"Invalid mode '{mode}'.")
    
    if args.enumeration is not None or (args.read is not None and args.orientation is not None):
        # in which case CSV and PLOT will always be generated
        args.csv = True
        args.plot = True
    
    # determining CSMs

    if args.enumeration is not None:
        print("\nMode: 'enumeration'")
        mu_max, rmss_max = args.enumeration
        mu_max = np.rint(mu_max).astype(int)
        if mu_max < 1: raise ValueError("Multiplicity should be a positive integer.")
        elif rmss_max <= 0: raise ValueError("Root-mean-square-strain should be a positive float.")
        if mu_max >= 8 or rmss_max > 0.5:
            print(f"Warning: Current MAX_MU = {mu_max:d} and MAX_RMSS = {rmss_max:.2f} may result in a large number of CSMs, which may take "
                    + "a long time to enumerate. If you are new to 'crystmatch', we recommend using MAX_MU <= 4 and MAX_RMSS = 0.4 first, "
                    + "and then increase MAX_MU and decrease MAX_RMSS gradually.")
        if args.initial == None: args.initial = input("Enter the path of the initial POSCAR file: ")
        crystA = load_poscar(args.initial, tol=args.tolerance)
        if args.final == None: args.final = input("Enter the path of the final POSCAR file: ")
        crystB = load_poscar(args.final, tol=args.tolerance)
        check_chem_comp(crystA[1], crystB[1])
        job = f"m{mu_max:d}s{rmss_max:.2f}"
        csm_bins, slmlist, mulist, rmsslist, rmsdlist = enum_rep(crystA, crystB, mu_max, rmss_max, args.tolerance)
        table = np.array([np.arange(len(mulist)), np.arange(len(mulist)), mulist, rmsslist, rmsdlist], dtype=float).T

    elif args.read is not None:
        print(f"\nMode: 'read'")
        data = np.load(args.read, allow_pickle=True)
        table = data['table']
        slmlist = data['slmlist']
        job, crystA, crystB, mu_max, rmss_max, symprec = data['metadata']
        print(f"Metadata from '{args.read}':\n\ttotal number of CSMs: {table.shape[0]:d}\n\tmultipliticy <= {mu_max}"
                + f"\n\trms strain <= {rmss_max}\n\tsymmetry tolerance: {symprec}")
        print(save_poscar(None, crystA, crystname="\nPrimitive cell of the initial structure (in POSCAR format):"))
        print(save_poscar(None, crystB, crystname="\nPrimitive cell of the final structure (in POSCAR format):"))

    elif args.single is not None:
        print("\nMode: 'single-CSM'")
        if args.initial == None: args.initial = input("Enter the path of the initial POSCAR file: ")
        crystA_sup = load_poscar(args.initial, tol=args.tolerance, to_primitive=False)
        if args.final == None: args.final = input("Enter the path of the final POSCAR file: ")
        crystB_sup = load_poscar(args.final, tol=args.tolerance, to_primitive=False)
        if not crystA_sup[1].shape[0] == crystB_sup[1].shape[0]: raise ValueError("\nNumbers of atoms in two crysts are not equal!")
        if not (crystA_sup[1] == crystB_sup[1]).all(): raise ValueError("\nAtom species in two crysts do not match!")
        crystA = load_poscar(args.initial, tol=args.tolerance, to_primitive=True, verbose=False)
        crystB = load_poscar(args.final, tol=args.tolerance, to_primitive=True, verbose=False)
        print(save_poscar(None, crystA, crystname="\nPrimitive cell of the initial structure (in POSCAR format):"))
        print(save_poscar(None, crystB, crystname="\nPrimitive cell of the final structure (in POSCAR format):"))
        job = "single"
        
        # single-CSM analysis
        print(f"\nGeometric properties of the CSM:")
        cA = crystA[0].T
        cB = crystB[0].T
        cA_sup = crystA_sup[0].T
        cB_sup = crystB_sup[0].T
        # computing SLM
        mA = (la.inv(cA) @ cA_sup).round().astype(int)
        mB = (la.inv(cB) @ cB_sup).round().astype(int)
        hA, qA = hnf_int(mA)
        hB, qB = hnf_int(mB)
        slmlist = [cong_slm((hA, hB, qB @ la.inv(qA).round().astype(int)),
                            get_pure_rotation(crystA, tol=args.tolerance), get_pure_rotation(crystB, tol=args.tolerance))[0]]
        mulist = multiplicity(crystA, crystB, slmlist)
        # computing principal strains
        _, sigma, vT = la.svd(cB_sup @ la.inv(cA_sup))
        rmsslist = [rmss(sigma)]
        # computing rmsd and optimized final structure
        cB_sup_final = vT.T @ np.diag(sigma) @ vT @ cA_sup
        c_sup_half = vT.T @ np.diag(sigma ** 0.5) @ vT @ cA_sup
        pA_sup = crystA_sup[2].T
        pB_sup = crystB_sup[2].T
        t0 = brute(lambda z: minimize_rmsd_tp(c_sup_half, pA_sup, pB_sup + z.reshape(3,1))[0],
            ((0,1),(0,1),(0,1)), Ns=10, finish=None)
        _, ks = minimize_rmsd_tp(c_sup_half, pA_sup, pB_sup + t0.reshape(3,1))
        t0 = (pA_sup - pB_sup - ks).mean(axis=1, keepdims=True)
        pB_sup_final = pB_sup + ks + t0
        rmsdlist = [la.norm(c_sup_half @ (pA_sup - pB_sup_final)) / np.sqrt(pA_sup.shape[1])]
        table = np.array([[0], [0], mulist, rmsslist, rmsdlist], dtype=float).T
        print(f"\tmultiplicity ∈ {{{','.join(str(mu) for mu in range(1,mulist[0]+1) if mulist[0] % mu == 0)}}}\n\tprincipal strains: "
                + ", ".join([f"{100*(s-1):.2f}%" for s in sigma]) + f"\n\trmss = {rmsslist[0]:.4f}\n\trmsd = {rmsdlist[0]:.4f} Å.")
        print(f"\nTo produce a list of CSMs that may contain this CSM, you can run:\n"
                + f"\n\t$ crystmatch -E {mulist[0]} {rmsslist[0]+0.005:.3f} -I '{args.initial}' -F '{args.final}'")

    print("")

    # orientation analysis
    
    if args.orientation != None:
        vix, viy, viz, vfx, vfy, vfz, wix, wiy, wiz, wfx, wfy, wfz = args.orientation
        print(f"Comparing each CSM with the given OR:\n\t[{vix},{viy},{viz}] || [{vfx},{vfy},{vfz}]"
                + f"\n\t[{wix},{wiy},{wiz}] || [{wfx},{wfy},{wfz}]")
        r = orient_matrix([vix,viy,viz], [vfx,vfy,vfz], [wix,wiy,wiz], [wfx,wfy,wfz])
        anglelist = deviation_angle(crystA, crystB, slmlist, r, uspfix=args.uspfix)
        table = np.hstack((table, anglelist.reshape(-1,1)))
        if args.single is not None: print(f"The deviation angle is {anglelist[0]:.6f} = {anglelist[0]*180/np.pi:.4f}°\n")

    # saving NPZ and POSCAR files
    
    if args.enumeration is not None:
        # saving enumeration results in NPZ format
        metadata = np.array([job, crystA, crystB, mu_max, rmss_max, args.tolerance], dtype=object)
        np.savez(unique_filename("Saving CSMs to", f"CSM_LIST-{job}.npz"), *csm_bins, metadata=metadata, slmlist=slmlist, table=table)
    
    if args.single is not None:
        # saving primitive cells
        dirname = unique_filename(f"Saving optimized final structure (rotation-free, translated to minimize RMSD) to", "CSM_single")
        makedirs(dirname)
        save_poscar(f"{dirname}{sep}{args.initial.split(sep)[-1]}", crystA_sup)
        save_poscar(f"{dirname}{sep}{splitext(args.final)[0].split(sep)[-1]}-optimized{splitext(args.final)[1]}", (cB_sup_final.T, crystB_sup[1], pB_sup_final.T))
    
    if args.export is not None:
        if args.read is None:
            print("\nValueWarning: '--export' option is only available in 'read' mode.")
        else:
            print(f"\nExporting CSMs with indices {', '.join(map(str, args.export))} in {args.read} ...")
            for i in args.export:
                if i >= table.shape[0]:
                    print(f"\nIndexWarning: Index {i} is out of range (there are only {table.shape[0]} CSMs).")
                    continue
                slm = data['slmlist'][table[i,1].round().astype(int)]
                mu = table[i,2].round().astype(int)
                ind = np.sum(table[:i,2].round().astype(int) == mu)
                p = data[f'arr_{mu:d}'][ind,:,0]
                ks = data[f'arr_{mu:d}'][ind,:,1:].T
                crystA_sup, crystB_sup = int_arrays_to_pair(crystA, crystB, slm, p, ks)
                dirname = unique_filename(f"Saving the CSM with index {i:d} to", f"CSM_{i:d}")
                makedirs(dirname)
                save_poscar(f"{dirname}{sep}POSCAR_I", crystA_sup, crystname=f"CSM_{i:d} initial")
                save_poscar(f"{dirname}{sep}POSCAR_F", crystB_sup, crystname=f"CSM_{i:d} final")
    
    # saving CSV file and plotting

    if args.csv:
        if args.orientation is None:
            np.savetxt(unique_filename("Saving results to", f"TABLE-{job}.csv"), table * np.array([1,1,1,100,1]), fmt='%8d,%8d,%8d,%8.4f,%8.4f',
                        header=' index,  slm_id,      mu,  rmss/%,  rmsd/Å')
        else:
            np.savetxt(unique_filename("Saving results to", f"TABLE-OR-{job}.csv"), table * np.array([1,1,1,100,1,180/np.pi]), fmt='%8d,%8d,%8d,%8.4f,%8.4f,%8.4f',
                        header=' index,  slm_id,      mu,  rmss/%,  rmsd/Å, theta/°')
    
    if args.plot:
        save_scatter(unique_filename("Creating scatter plot in", f"PLOT-{job}.pdf"), table[:,3], table[:,4], table[:,2].round().astype(int), cbarlabel="Multiplicity")
        if args.orientation is not None: save_scatter(unique_filename("Creating scatter plot in", f"OR-{job}.pdf"),
                                                        table[:,3], table[:,4], table[:,5], cbarlabel="Deviation / °", cmap=plt.cm.get_cmap('jet'))
    
    print(f"\nTotal time spent: {time()-time0:.2f} seconds")

if __name__ == "__main__":
    main()


import os
import re
import setuptools

def parse_init_file():
    init_path = os.path.join(os.path.dirname(__file__), 'crystmatch', '__init__.py')
    variables = {}
    pattern = re.compile(r'^__(\w+)__\s*=\s*[\'"]([^\'"]+)[\'"]')
    with open(init_path, 'r', encoding='utf-8') as f:
        for line in f:
            match = pattern.match(line.strip())
            if match:
                key, value = match.groups()
                variables[key] = value
    return variables

metadata = parse_init_file()

with open("README.md", "r", encoding="utf-8") as fhand:
    long_description = fhand.read()

setuptools.setup(
    name=metadata["name"],
    version=metadata["version"],
    author=metadata["author"],
    author_email=metadata["email"],
    description=(metadata["description"]),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=metadata["url"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=["numpy", "scipy", "matplotlib", "spglib"],
    packages=setuptools.find_packages(),
    python_requires=">=3.9",
    entry_points={
        "console_scripts": [
            "crystmatch = crystmatch:main",
        ]
    }
)